
# Thompson Variable Selection for large streaming dataset

TVS_stream<-function(Y, X, selector ='bart', topq, niter= 1000, ntree = 10,
                     a = 1, b =1 , binary_reward = TRUE, batchsize,
                     fix_stop, maxround = 100, stop_crit =100){
  #' TVS_stream
  #' 'Online' TVS which converges only after fix number of rounds \strong{and} the convergence criteria has been met
  #' @param Y The response, have to be continuous
  #' @param X The matrix of variables
  #' @param selector support BART selector 'bart' or SSLASSO selector 'SSLASSO'
  #' @param topq An optional system. selecting variable with posterior probability greater than 0.5, we have top q variables
  #' @param stop_crit There is a stopping critria if the model has not changed for `stop_crit` iterations, the algorithm is deemed to reach convergence
  #' @param fix_stop If `fix_stop` is not missing, then the the algorithm stops after fixed number of stops and the stop_crit is met
  #' @param niter The number of MCMC iterations in BART
  #' @param ntree The number fo trees in BART
  #' @param a initial prior a
  #' @param b initial prior b
  #' @param binary_reward Whether to binarize the reward or choose a reward based on a bernoulli flip
  #' @param batchsize The size of each batch
  #' @param maxround the max number of round to go through the data
  #' @return A class 'TVS' object containing the following:
  #' \itemize{
  #' \item `A`, `B` The TVS A and B each row is `a_i` and `b_i` across time
  #' \item `iter_converge` The iteration number when the algorithm reach convergence
  #' \item `round_converge` The round number when the algorithm reach convergence
  #' \item `iter_fixed` The iteration number when the algorithm reach convergence
  #' \item `time_fixed` Time taken for the algorithm to run fix number of rounds if fix_stop is missing, it is time taken to run one round
  #' \item `time_converge` Time taken for the algorithm to reach convergence
  #' \item `type` either 'offline' or 'online' the type of algorithm used, in this case 'online'
  #' }
  #' @export
  #' @examples
  #' set.seed(2020)
  #' X <- matrix(rnorm(1000000),10000,100)
  #' Y <- 2 + X[,1]+ 3*X[,2] + rnorm(10000)
  #' TVS_output <- TVS_stream(Y,X,selector= 'bart', niter= 1000, ntree = 10,
  #' a = 1, b =1 , binary_reward = TRUE, batchsize = 500,
  #' fix_stop = 5, maxround = 100, stop_crit =100)
  #' posterior(TVS_output)


  #list(A=A,B=B, iter_converge= (iter_converge-1), round_converge = k_converge, iter_fixed = (iter_fixed-1),
  #    time_fixed = time_fixed, time_converge = converge_time,  type = 'online')



  if(!missing(fix_stop)){ #if the user want to have a fix number of iteration included, then he/she should have fix_stop
    #The fix stop will be round.

    stopifnot(fix_stop <= maxround)

    round <- fix_stop

  }else{ #if not, the round will be 1 and since convergence round is at least 1, we are done.

    round <- 1
  }


  converge_iter <- stop_crit #the number of iteration for which the model do not change is initially called converge_iter,
                             #then for consistency with TVS, it is renamed stop_crit so this is done to have least demage on the code

  start_time = Sys.time() #start timing

	p<-ncol(X)

	n<-nrow(X)

	nrep<-floor(n/batchsize)

	A<-matrix(a,p,nrep*maxround)

	B<-matrix(b,p,nrep*maxround)

	Aprev<-numeric(p)

	Anew<-numeric(p)

	Bprev<-numeric(p)

	Bnew<-numeric(p)

	con_criteria <- rep(1,converge_iter)

	iter <- 1

	#Because we need to end it at the round than at iteration itself and I want to compute both converge iter and fix iter
	#at the same time, I use to

	converged <- FALSE # converge is FALSE

	First_convergence <- TRUE #It is true because the convergence has not been met, once it has converged, it will be FALSE
	#This stop the system from writing over results if fix_round (fix_k) > converge_round (converge_k)

	for (k in (1:maxround)){

		permute<-sample(1:n)

		Yperm<-Y[permute]

		Xperm<-X[permute,]

	  for (i in (1:nrep)){

		  Aprev<-A[,(k-1)*nrep+i]

		  Bprev<-B[,(k-1)*nrep+i]

		  thetas<-rbeta(p,Aprev,Bprev)

		  if (missing(topq)){

			  gammas<-as.numeric(thetas> 0.5)

			}
		  else{

			  gammas<-numeric(p)

			  index<-order(thetas,decreasing=T)

			  gammas[index[1:topq]]<-1

			}

		  # Collect reward

		  index<-(i-1)*batchsize+(1:batchsize)

		  Ysub<-Yperm[index]

		  Xsub<-Xperm[index,]

		  rbin<-reward(Ysub,Xsub,gammas,type=selector,
					 niter,ntree=ntree,binary_reward=binary_reward)

		  if(max(rbin)>0){

		    rbin<-rbinom(sum(gammas),1,rbin/max(rbin))

			}

		  # Update

		  Anew<-Aprev

		  Anew[gammas==1]<-Anew[gammas==1]+rbin

		  Bnew<-Bprev

		  Bnew[gammas==1]<-Bnew[gammas==1]+1-rbin

		  #record convergence

		  con_criteria[(iter%%converge_iter)+1] <- sum(abs((Anew/(Anew+Bnew)>0.5)-(Aprev/(Aprev+Bprev)>0.5)))


		  if(sum(con_criteria) ==0){

		    converged = TRUE

		  }

		  iter = iter + 1

		  if(i < nrep | k< maxround){

		    A[,(k-1)*nrep+i+1]<-Anew

		    B[,(k-1)*nrep+i+1]<-Bnew

		  }

	  }

		#check for convergence condition.

		if(converged){

		  if(First_convergence){ #this is when the convergence is meet for the first time. If there is no first
		    #convergence, the convergence can be met but fix_k > converge_k so each time I pass a round,
		    #the result can be written over and therefore the result is not desirable.

		    First_convergence = FALSE #This is to avoid the result being written over when fix_k > converge_k

		    iter_converge <- iter

		    k_converge = k

		    converge_time =Sys.time() - start_time # this record the number of rounds

		  }

		  if(k > round){

		    break #converge_k > fix_k, then this condition kicks in and we are done.

		  }

		}
		if(k == round){ #This is when system has done fix_k number of time.

		  iter_fixed = iter

		  time_fixed = Sys.time() - start_time

		  if(converged){ #converge_k <= fix_k

		    break

		  }

		}

	}
	#The iteration has reach maxround at this point if it did not break.
	if(k == maxround){ #This is when algorithm never converged

	  iter_fixed = iter

	  k_converge = k

	  converge_time =Sys.time() - start_time # this record the number of rounds

	}
	if(First_convergence){ #It is to ensure that

	  #This is when algorithm never converged

	  First_convergence = FALSE

	}

	r <- list(A=A,B=B, iter_converge= (iter_converge-1), round_converge = k_converge, iter_fixed = (iter_fixed-1),
	          time_fixed = time_fixed, time_converge = converge_time,  type = 'online')

  class(r) <- 'TVS_stream'

  return(r)
}
